/*
 * base64Utility.c
 * Created by M.Miyazaki [2013.01.01]
 * Modified by M.Miyazaki [2013.01.02]
 */

#include "base64Utility.h"


/**
 * Printable characters.
 */
const char *printables = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/**
 * Padding character.
 */
const char padding = '=';


/**
 * Encode to Base64 string.
 */
char* encodeBase64(const char *aString)
{
	int stringLength, index1, index2 = 0;
	char *encodedString;
	
	stringLength = strlen(aString);
	encodedString = (char *)malloc(sizeof(char) * ((((stringLength - 1) / 3) + 1) * 4 + 1));
	if (encodedString == NULL)
	{
		return NULL;
	}
	if (stringLength == 0)
	{
		return encodedString;
	}
	
	for (index1 = 0; index1 + 3 <= stringLength; index1 += 3)
	{
		encodedString[index2++] = printables[(aString[index1] & 0xFC) >> 2];
		encodedString[index2++] = printables[((aString[index1] & 0x03) << 4) | ((aString[index1+1] & 0xF0) >> 4)];
		encodedString[index2++] = printables[((aString[index1+1] & 0x0F) << 2) | ((aString[index1+2] & 0xC0) >> 6)];
		encodedString[index2++] = printables[aString[index1+2] & 0x3F];
	}
	
	if (index1 + 1 == stringLength)
	{
		encodedString[index2++] = printables[(aString[index1] & 0xFC) >> 2];
		encodedString[index2++] = printables[(aString[index1] & 0x03) << 4];
		encodedString[index2++] = padding;
		encodedString[index2++] = padding;
	}
	if (index1 + 2 == stringLength)
	{
		encodedString[index2++] = printables[(aString[index1] & 0xFC) >> 2];
		encodedString[index2++] = printables[((aString[index1] & 0x03) << 4) | ((aString[index1+1] & 0xF0) >> 4)];
		encodedString[index2++] = printables[(aString[index1+1] & 0x0F) << 2];
		encodedString[index2++] = padding;
	}
	encodedString[index2] = '\0';
	
	return (char *)encodedString;
}

/**
 * Decode from Base64 string.
 */
char* decodeBase64(const char *base64String)
{
	int stringLength, index1, index2 = 0;
	int firstValue, secondValue, thirdValue, fourValue;
	char *decodedString;
	
	stringLength = strlen(base64String);
	decodedString = (char *)malloc(sizeof(unsigned char) * (stringLength / 4 * 3 + 1));
	if (decodedString == NULL)
	{
		return NULL;
	}
	if (stringLength == 0)
	{
		return decodedString;
	}
	
	for (index1 = 0; index1 + 4 < stringLength; index1 += 4)
	{
		firstValue  = indexOf(base64String[index1]);
		secondValue = indexOf(base64String[index1+1]);
		thirdValue  = indexOf(base64String[index1+2]);
		fourValue   = indexOf(base64String[index1+3]);
		decodedString[index2++] = (firstValue << 2) | ((secondValue & 0x30) >> 4);
		decodedString[index2++] = ((secondValue & 0x0F) << 4) | ((thirdValue & 0x3C) >> 2);
		decodedString[index2++] = ((thirdValue & 0x03) << 6) | (fourValue & 0x3F);
	}
	
	firstValue  = indexOf(base64String[index1]);
	secondValue = indexOf(base64String[index1+1]);
	thirdValue  = indexOf(base64String[index1+2]);
	fourValue   = indexOf(base64String[index1+3]);
	decodedString[index2++] = (firstValue << 2) | ((secondValue & 0x30) >> 4);
	if (thirdValue != -1)
	{
		decodedString[index2++] = ((secondValue & 0x0F) << 4) | ((thirdValue & 0x3C) >> 2);
	}
	if (fourValue != -1)
	{
		decodedString[index2++] = ((thirdValue & 0x03) << 6) | (fourValue & 0x3F);
	}
	decodedString[index2] = '\0';
	
	return (char *)decodedString;
}

/**
 * Search printable character and return the index.
 */
int indexOf(char printableCharacter)
{
	int index, returnIndex = -1;
	
	for (index = 0; index < 64; index++)
	{
		if (printables[index] == printableCharacter)
		{
			returnIndex = index;
			break;
		}
	}
	
	return returnIndex;
}
