%{

%}

%option nounput

/*** Lexical elements ***/
Token					({Keyword}|{Identifier}|{Constant}|{StringLiteral}|{Punctuator})
PreprocessingToken		({HeaderName}|{Identifier}|{PpNumber}|{CharacterConstant}|{StringLiteral}|{Punctuator})

/*** Keywords ***/
Keyword		("auto"|"break"|"case"|"char"|"const"|"continue"|"default"|"do"|"double"|"else"|"enum"|"extern"|"float"|"for"|"goto"|"if"|"inline"|"int"|"long"|"register"|"restrict"|"return"|"short"|"signed"|"sizeof"|"static"|"struct"|"switch"|"typedef"|"union"|"unsigned"|"void"|"volatile"|"while"|"_Bool"|"_Complex"|"_Imaginary")

/*** Identifiers ***/
Identifier				({IdentifierNonDigit}({IdentifierNonDigit}|{Digit})*)
IdentifierNonDigit		({NonDigit}|{UniversalCharacterName})
Digit					[0-9]
NonDigit				[_a-zA-Z]

/*** Universal character names ***/
UniversalCharacterName		(\\u{HexQuad}|\\U{HexQuad}{HexQuad})
HexQuad						({HexadecimalDigit}{4})

/*** Constants ***/
IntegerConstant					(({DecimalConstant}|{OctalConstant}|{HexadecimalConstant}){IntegerSuffix}?)
DecimalConstant					({NonZeroDigit}{Digit}*)
OctalConstant					(0{OctalDigit}*)
HexadecimalConstant				({HexadecimalPrefix}{HexadecimalDigit}+)
HexadecimalPrefix				(0[xX])
NonZeroDigit					[1-9]
OctalDigit						[0-7]
HexadecimalDigit				[0-9a-fA-F]
IntegerSuffix					({UnsignedSuffix}({LongSuffix}?|{LongLongSuffix})|({LongSuffix}|{LongLongSuffix}){UnsignedSuffix}?)
UnsignedSuffix					[uU]
LongSuffix						[lL]
LongLongSuffix					("ll"|"LL")
FloatingConstant				({DecimalFloatingConstant}|{HexadecimalFloatingConstant})
DecimalFloatingConstant			(({FractionalConstant}{ExponentPart}?|{DigitSequence}{ExponentPart}){FloatingSuffix}?)
HexadecimalFloatingConstant		({HexadecimalPrefix}({HexadecimalFractionalConstant}|{HexadecimalDigitSequence}){BinaryExponentPart}{FloatingSuffix}?)
FractionalConstant				({DigitSequence}?"."{DigitSequence}|{DigitSequence}".")
ExponentPart					([eE]{Sign}?{DigitSequence})
Sign							("+"|"-")
DigitSequence					({Digit}+)
HexadecimalFractionalConstant	({HexadecimalDigitSequence}?"."{HexadecimalDigitSequence}|{HexadecimalDigitSequence}".")
BinaryExponentPart				([pP]{Sign}?{DigitSequence})
HexadecimalDigitSequence		({HexadecimalDigit}+)
FloatingSuffix					[flFL]
EnumerationConstant				{Identifier}
CharacterConstant				(L?'{CCharSequence}')
CCharSequence					({CChar}+)
CChar							([^'\\\n\r]|{EscapeSequence})
EscapeSequence					({SimpleEscapeSequence}|{OctalEscapeSequence}|{HexadecimalEscapeSequence}|{UniversalCharacterName})
SimpleEscapeSequence			(\\['"?\\abfnrtv])
OctalEscapeSequence				(\\{OctalDigit}{1,3})
HexadecimalEscapeSequence		(\\x{HexadecimalDigit}+)

/*** String literals ***/
StringLiteral		L?\"{SCharSequence}?\"
SCharSequence		({SChar}+)
SChar				([^\"\\\n\r]|{EscapeSequence})

/*** Punctuators ***/
Punctuator		("["|"]"|"("|")"|"{"|"}"|"."|"->"|"++"|"--"|"&"|"*"|"+"|"-"|"~"|"!"|"/"|"%"|"<<"|">>"|"<"|">"|"<="|">="|"=="|"!="|"^"|"|"|"&&"|"||"|"?"|":"|";"|"..."|"="|"*="|"/="|"%="|"+="|"-="|"<<="|">>="|"&="|"^="|"|="|","|"#"|"##"|"<:"|":>"|"<%"|"%>"|"%:"|"%:%:")

/*** Header names ***/
HeaderName			("<"{HCharSequence}">"|\"{QCharSequence}\")
HCharSequence		({HChar}+)
HChar				[^\n\r>]
QCharSequence		({QChar}+)
QChar				[^\n\r\"]

/*** Preprocessing numbers ***/
PpNumber		("."?{Digit}({Digit}|{IdentifierNonDigit}|[eEpP]{Sign}|".")*)

/*** Preprocessing directives ***/
PreprocessingFile		({Group}?)
Group					({GroupPart}+)
GroupPart				({IfSection}|{ControlLine}|{TextLine}|"#"{NonDirective})
IfSection				({IfGroup}{ElifGroups}?{ElseGroup}?{EndifLine})
IfGroup					("#"("if"{ConstantExpression}|("ifdef"|"ifndef"){Identifier}){NewLine}{Group}?)
ElifGroups				({ElifGroup}+)
ElifGroup				("#""elif"{ConstantExpression}{NewLine}{Group}?)
ElseGroup				("#""else"{NewLine}{Group}?)
EndifLine				("#""endif"{NewLine})
ControlLine				("#"())

/*
 Excepted Rule.
 {EnumerationConstant}		{ printf("# EnumerationConstant : %s\n", yytext); return(CONSTANT); }
*/


%%
"/*"								{ token(); comment(); }
"//"[^\r\n]*("\r\n"|"\r"|"\n")		{ token(); commentLines++; physicalLOC++; newLine();  /*  One line comment */ }

"auto"			{ token(); effectiveToken(); return(AUTO); }
"register"		{ token(); effectiveToken(); return(REGISTER); }
"static"		{ token(); effectiveToken(); return(STATIC); }
"extern"		{ token(); effectiveToken(); return(EXTERN); }
"typedef"		{ token(); effectiveToken(); return(TYPEDEF); }
"void"			{ token(); effectiveToken(); return(VOID); }
"char"			{ token(); effectiveToken(); return(CHAR); }
"short"			{ token(); effectiveToken(); return(SHORT); }
"int"			{ token(); effectiveToken(); return(INT); }
"long"			{ token(); effectiveToken(); return(LONG); }
"float"			{ token(); effectiveToken(); return(FLOAT); }
"double"		{ token(); effectiveToken(); return(DOUBLE); }
"signed"		{ token(); effectiveToken(); return(SIGNED); }
"unsigned"		{ token(); effectiveToken(); return(UNSIGNED); }
"_Bool"			{ token(); effectiveToken(); return(BOOL); }
"_Complex"		{ token(); effectiveToken(); return(COMPLEX); }
"const"			{ token(); effectiveToken(); return(CONST); }
"restrict"		{ token(); effectiveToken(); return(RESTRICT); }
"volatile"		{ token(); effectiveToken(); return(VOLATILE); }
"inline"		{ token(); effectiveToken(); return(INLINE); }
"struct"		{ token(); effectiveToken(); return(STRUCT); }
"union"			{ token(); effectiveToken(); return(UNION); }
"enum"			{ token(); effectiveToken(); return(ENUM); }
"case"			{ token(); effectiveToken(); return(CASE); }
"default"		{ token(); effectiveToken(); return(DEFAULT); }
"if"			{ token(); effectiveToken(); return(IF); }
"else"			{ token(); effectiveToken(); return(ELSE); }
"switch"		{ token(); effectiveToken(); return(SWITCH); }
"while"			{ token(); effectiveToken(); return(WHILE); }
"do"			{ token(); effectiveToken(); return(DO); }
"for"			{ token(); effectiveToken(); return(FOR); }
"goto"			{ token(); effectiveToken(); return(GOTO); }
"continue"		{ token(); effectiveToken(); return(CONTINUE); }
"break"			{ token(); effectiveToken(); return(BREAK); }
"return"		{ token(); effectiveToken(); return(RETURN); }
"sizeof"		{ token(); effectiveToken(); return(SIZEOF); }

{IntegerConstant}			{ token(); effectiveToken(); return(CONSTANT); }
{FloatingConstant}			{ token(); effectiveToken(); return(CONSTANT); }
{CharacterConstant}			{ token(); effectiveToken(); return(CONSTANT); }

{Identifier}		{ token(); effectiveToken(); return ((isTypedefName(yytext) == 1) ? TYPEDEF_NAME : IDENTIFIER); }
{StringLiteral}		{ token(); effectiveToken(); return(STRING_LITERAL); }

"..."		{ token(); effectiveToken(); return(ELLIPSIS); }
"<<="		{ token(); effectiveToken(); return(LEFT_SHIFT_ASSIGNMENT); }
">>="		{ token(); effectiveToken(); return(RIGHT_SHIFT_ASSIGNMENT); }
"*="		{ token(); effectiveToken(); return(MULTIPLICATION_ASSIGNMENT); }
"/="		{ token(); effectiveToken(); return(DIVISION_ASSIGNMENT); }
"%="		{ token(); effectiveToken(); return(REMAINDER_ASSIGNMENT); }
"+="		{ token(); effectiveToken(); return(ADDITION_ASSIGNMENT); }
"-="		{ token(); effectiveToken(); return(SUBTRACTION_ASSIGNMENT); }
"&="		{ token(); effectiveToken(); return(BITWISE_AND_ASSIGNMENT); }
"^="		{ token(); effectiveToken(); return(BITWISE_EXCLUSIVE_OR_ASSIGNMENT); }
"|="		{ token(); effectiveToken(); return(BITWISE_INCLUSIVE_OR_ASSIGNMENT); }
"=="		{ token(); effectiveToken(); return(EQUAL_OPERATOR); }
"!="		{ token(); effectiveToken(); return(NOT_EQUAL_OPERATOR); }
"<="		{ token(); effectiveToken(); return(LESS_THAN_EQUAL_OPERATOR); }
">="		{ token(); effectiveToken(); return(GREATER_THAN_EQUAL_OPERATOR); }
"<<"		{ token(); effectiveToken(); return(LEFT_SHIFT_OPERATOR); }
">>"		{ token(); effectiveToken(); return(RIGHT_SHIFT_OPERATOR); }
"&&"		{ token(); effectiveToken(); return(LOGICAL_AND); }
"||"		{ token(); effectiveToken(); return(LOGICAL_OR); }
"++"		{ token(); effectiveToken(); return(PLUS_PLUS); }
"--"		{ token(); effectiveToken(); return(MINUS_MINUS); }
"->"		{ token(); effectiveToken(); return(ARROW_OPERATOR); }
"<"			{ token(); effectiveToken(); return(LESS_THAN_OPERATOR); }
">"			{ token(); effectiveToken(); return(GREATER_THAN_OPERATOR); }
"{"			{ token(); braceOrParen(); indentLevel++; return(LEFT_BRACE); }
"}"			{ token(); indentLevel--; braceOrParen(); return(RIGHT_BRACE); }
"["			{ token(); effectiveToken(); return(LEFT_BRACKET); }
"]"			{ token(); effectiveToken(); return(RIGHT_BRACKET); }
"("			{ token(); braceOrParen(); indentLevel++; return(LEFT_PAREN); }
")"			{ token(); indentLevel--; braceOrParen(); return(RIGHT_PAREN); }
"&"			{ token(); effectiveToken(); return(AND); }
"^"			{ token(); effectiveToken(); return(EXCLUSIVE_OR); }
"|"			{ token(); effectiveToken(); return(INCLUSIVE_OR); }
"="			{ token(); effectiveToken(); return(EQUAL); }
"*"			{ token(); effectiveToken(); return(ASTERISK); }
"/"			{ token(); effectiveToken(); return(SLASH); }
"%"			{ token(); effectiveToken(); return(PERCENT); }
"+"			{ token(); effectiveToken(); return(PLUS); }
"-"			{ token(); effectiveToken(); return(MINUS); }
"~"			{ token(); effectiveToken(); return(TILDE); }
"!"			{ token(); effectiveToken(); return(EXCLAMATION_MARK); }
"?"			{ token(); effectiveToken(); return(QUESTION_MARK); }
":"			{ token(); effectiveToken(); return(COLON); }
";"			{ token(); effectiveToken(); return(SEMICOLON); }
","			{ token(); effectiveToken(); return(COMMA); }
"."			{ token(); effectiveToken(); return(PERIOD); }
"#"			{ token(); preprocessor_line(); }

"\r\n"|"\r"|"\n"		{ token(); physicalLOC++; newLine(); }
[ \t]		{ token(); space(); }
[\f\v]		{ token(); }

.			{ token(); return(UNKNOWN); }
%%

int yywrap(void)
{
	return 1;
}

void yyerror(const char *aString)
{
	// Nothing to do...
	
	printf("#%s : %s#\n", yytext, aString);
	
	return;
}


/**
 * For parse Preprocessor Line.
 */
void preprocessor_line(void)
{
	char current, previous;
	
	while((current = input()) != EOF)
	{
		if (debugMode == TRUE)
		{
			printf("%c", current);
		}
		
		if (current == '\r' || (previous != '\r' && current == '\n'))
		{
			physicalLOC++;
		}
		
		if (previous != '\\' && (current == '\n' || current == '\r'))
		{
			return;
		}
		
		previous = current;
	}
	
	yyerror("Unterminated preprocessor line.");
}

/**
 * For parse Multiline Comment.
 */
void comment(void)
{
	char current, previous;
	
	while ((current = input()) != EOF)
	{
		if (debugMode == TRUE)
		{
			printf("%c", current);
		}
		
		if (current == '\r' || (previous != '\r' && current == '\n'))
		{
			commentLines++;
			physicalLOC++;
		}
		
		if (previous == '*' && current == '/')
		{
			commentLines++;
			return;
		}
		
		previous = current;
	}
	
	yyerror("Unterminated comment.");
}

/**
 * Action for new-line character.
 */
void newLine(void)
{
	if (isEffectiveLine == TRUE)
	{
		effectiveLOC++;
		isEffectiveLine = FALSE;
	}
	
	if (isInIndent == FALSE)
	{
		isInIndent = TRUE;
	}
	
	indentSpaces = 0;
	indentTabs   = 0;
	
	return;
}

/**
 * Action for effective token.
 */
void effectiveToken(void)
{
	if (isEffectiveLine == FALSE)
	{
		isEffectiveLine = TRUE;
	}
	
	if (isInIndent == TRUE)
	{
		isInIndent = FALSE;
		if (debugMode == TRUE) printf("[## indentLevel : %d, space : %d, tab : %d ##]", indentLevel, indentSpaces, indentTabs);
		registerIndent(indentLevel, indentSpaces, indentTabs);
	}
	
	return;
}

/**
 * Action for space.
 */
void space(void)
{
	if (isEffectiveLine == FALSE)
	{
		if (strcmp(yytext, " ") == 0)
		{
			indentSpaces++;
		}
		if (strcmp(yytext, "\t") == 0)
		{
			indentTabs++;
		}
	}
	
	return;
}

/**
 * Action for token.
 */
void token(void)
{
	if (debugMode == TRUE)
	{
		printf("%s", yytext);
	}
	
	return;
}

/**
 * Action for brace or parenthesis.
 */
void braceOrParen()
{
	if (isInIndent == TRUE)
	{
		isInIndent = FALSE;
		if (debugMode == TRUE) printf("[## indentLevel : %d, space : %d, tab : %d ##]", indentLevel, indentSpaces, indentTabs);
		registerIndent(indentLevel, indentSpaces, indentTabs);
	}
	
	return;
}
